// Copyright (c) 2000-2013 ActiveState Software Inc.
// See the file LICENSE.txt for licensing information.
//---- globals

var { classes: Cc, interfaces: Ci, utils: Cu } = Components;

if (typeof(ko)=='undefined') {
    var ko = {};
}
if (typeof(ko.refactoring)=='undefined') {
    ko.refactoring = {};
}
if (typeof(ko.refactoring.renameAttribute)=='undefined') {
    ko.refactoring.renameAttribute = {};
}

var widgets = {};
var log = ko.logging.getLogger("refactoring.confirmRenameAttributeChange");
//log.setLevel(ko.logging.LOG_DEBUG);
var g_args;
var g_opener;
var g_defn;
var g_displayTree = null;

var hitsByURL; /* { url => {
                     path => path,
                     hitListLines => { lineNum => { context: line of text,
                                                    hitList: [startIndex, endIndex,
                                                              columnNo] } } } } */

var filteredHitsByURL; // same as above 

var scopeStart;
var scopeEnd;

function init_widgets () {
    widgets.repls = document.getElementById("repls");
    widgets.showChangesButton = document.getElementById("show-marked-changes-btn");
};

function init() {
    g_args = window.arguments[0];
    g_opener = g_args.opener;
    g_defn = g_args.defn;
    scopeStart = g_defn.scopestart;
    scopeEnd = g_defn.scopeend;
}

// hits have the form [startPos, endPos, column]
// startPos and endPos are in terms of the document, so they
// need to be normalized.
// column is one-based, so it needs to be adjusted by -1
function LineManager(textLine, repl, hits) {
    this.textLine = textLine;
    this.repl = repl;
    this.parts = [];
    let lim = hits.length;
    this.origs = new Array(lim);
    this.actives = new Array(lim);
    let pos = 0;
    for (let i = 0; i < lim; i++) {
        let hit = hits[i];
        let hitStart = hit[2] - 1;
        let startPos = hit[0], endPos = hit[1];
        let hitEnd = hitStart + endPos - startPos;
        this.parts.push(textLine.substring(pos, hitStart));
        pos = hitEnd;
        let orig = textLine.substring(hitStart, hitEnd)
        this.origs[i] = orig;
        this.actives[i] = orig;
    }
    this.parts.push(textLine.substr(pos));
    this.textChangedSinceLastRead = false;
};

LineManager.prototype = {
 applyChange: function applyChange(i) {
        this.actives[i] = this.repl;
        this.textChangedSinceLastRead = true;
    },
 undoChange: function undoChange(i) {
        this.actives[i] = this.origs[i];
        this.textChangedSinceLastRead = true;
    },
 undoAll: function undoAll() {
        for (let i = 0; i < this.actives.length; i++) {
            this.actives[i] = this.origs[i];
        }
        this.textChangedSinceLastRead = true;
    },
 applyAll: function undoAll() {
        for (let i = 0; i < this.actives.length; i++) {
            this.actives[i] = this.repl;
        }
        this.textChangedSinceLastRead = true;
    },
 getText: function getText() {
        let lim = this.parts.length;
        let newParts = new Array(lim * 2 + 1);
        lim -= 1;
        let j = 0;
        for (let i = 0; i < lim; i++) {
            newParts[j++] = this.parts[i];
            newParts[j++] = this.actives[i];
        }
        newParts[j] = this.parts[lim];
        this.textChangedSinceLastRead = false;
        return newParts.join("");
    }
};
            
function onLoad () {
    try {
        init_widgets();
        init();
        startSearch();
    } catch(ex) {
        log.exception(ex, "confirmRenameAttributeChanges.js: onLoad. Error");
    }
};

function cancel() {
    return true;
}

function doHits(hitsByURL) {
    g_displayTree = new TreeView();
    g_displayTree.updateRows(hitsByURL);
    widgets.view_repls = g_displayTree;
    widgets.repls.treeBoxObject.view = widgets.view_repls;
    for (let i = g_displayTree.rowCount - 1; i >= 0; --i) {
        g_displayTree._wrapToggleOpenState(i);
    }
    // And update the text of each row.  Otherwise if the user presses
    // "Show Changes" first any rows below the last visible line
    // won't be updated.
    for (let i = g_displayTree.rowCount - 1; i >= 0; --i) {
        g_displayTree.getCellText(i, {id: "text"});
    }
    
    updateUI('g_displayTree');
}

function onTreeClick(aEvent) {
    if (aEvent.button == 2) {
        return;
    }
    var row = {}, col = {};
    widgets.repls.treeBoxObject.getCellAt(aEvent.clientX, aEvent.clientY,
                                    row, col, {});
    row = row.value;
    col = col.value;
    if (!col) {
        return;
    }
    if (aEvent.metaKey || aEvent.ctrlKey || aEvent.shiftKey || aEvent.altKey) {
        return;
    }
    if (col.id == "isChecked") {
        try {
            if (g_displayTree._invoked_toggleOpenState) {
                // We processed an open/close container event
                g_displayTree._invoked_toggleOpenState = false;
                return;
            }
            try {
                g_displayTree.toggleRowChecked(row);
            } catch(ex) {
                log.exception(ex, "Prob in toggleRowChecked");
            }
        } catch(ex) {
            log.exception("toggleRowChecked => blew up");
        }
    }
}

function onTreeDblClick(aEvent) {
    // No post-processing needed.  Use this to handle
    // tree toggling, no need to guard other event-related processing,
    // such as checkbox toggling.
    aEvent.stopPropagation();
    aEvent.preventDefault();
}

function TreeView() {
   xtk.hierarchicalTreeView.apply(this, []);
   this._invoked_toggleOpenState = false;
};
TreeView.prototype = new xtk.hierarchicalTreeView();
TreeView.prototype.constructor = TreeView;

TreeView.prototype.updateRows = function updateRows(tableData) {
    var treeRows = [];
    var repl = g_args.repl;
    if (!repl) {
        throw new Error("Refactoring: RenameAttribute: No replacement specified");
    }
    var populateChildren = function populateChildren(targetA, hitListLines) {
        for (let lineNo in hitListLines) {
            let hit2 = hitListLines[lineNo];
            let context = hit2.context, hitList = hit2.hitList;
            let lineChangeMgr = new LineManager(context, repl, hitList);
            let inLineOccurrence = 0;
            hitList.forEach(function(hit3) {
                let childRow = {
                    isContainer: false,
                    level: 1,
                    lineChangeMgr: lineChangeMgr,
                    inLineOccurrence: inLineOccurrence++,
                    isChecked: true,
                    lineNo: lineNo
                };
                targetA.push(childRow);
            });
        }
    };
    for (let url in tableData) {
        let hit1 = tableData[url];
        let containerRow = {
            lineNo: "",
            path: hit1.path,
            isChecked: true,
            isContainer: true,
            hasChildren: true,
            state: xtk.hierarchicalTreeView.STATE_CLOSED,
            level: 0,
            children: [],
            getChildren: function() {
                return this.children
            }
        };
        populateChildren(containerRow.children, hit1.hitListLines)
        treeRows.push(containerRow);
    }
    this.rows = treeRows;
};

TreeView.prototype.getCellProperties = function(rowNum, column) {
    var col_id = column.id
    var row = this._rows[rowNum];
    //dump("getCellProperties[" + rowNum + "," + col_id + "] => ");
    if (col_id == "isChecked") {
        if (row.isChecked) {
            //dump("[checked]\n");
            return ["checked"];
        } else {
            return [];
        }
    }
    var res = xtk.hierarchicalTreeView.prototype.getCellProperties.apply(this, [rowNum, column]);
    //dump(res + "\n");
    return res;
};

TreeView.prototype.cycleCell = function(rowNum, column) {
    //dump("cycleCell[" + rowNum + "," + col_id + "] => \n");
};

TreeView.prototype.isEditable = function(rowNum, column) {
    var col_id = column.id
    var row = this._rows[rowNum];
    return col_id == "isChecked";
};

TreeView.prototype.getCellText = function(rowNum, column) {
    //
    // Warning: getCellText(rowNum, {id:"text"}) strips out the
    // leading white-space, as it's intended for display only (and we
    // assume the leading white-space isn't interesting.  For the
    // actual text, call the cell's lineChangeMgr.getText() method.
    var col_id = column.id
    var s  = ">> getCellText:" + rowNum + ", " + col_id;
    //dump(s);
    var res;
    try {
        var row = this._rows[rowNum];
        switch(col_id) {
            case "isChecked":
            case "lineNo":
             //dump(" => " + row[col_id] + "\n");
                return row[col_id];
            case "text":
                // do this later
                break;
            default:
             //dump("Unexpected columnID of " + col_id);
                return "";
        }
        let res;
        if (row.isContainer) {
            res = row.path;
        } else {
            let { lineChangeMgr: lineChangeMgr, inLineOccurrence: inLineOccurrence,
                    isChecked:checked } = row;
            lineChangeMgr[checked ? "applyChange" : "undoChange"](inLineOccurrence);
            res = lineChangeMgr.getText().replace(/^[ \t]+/, "");
        }
        //dump(" => " + res + "\n");
        return res;
    } catch(ex) {
        log.exception(ex, s);
        return ""
    }
};

TreeView.prototype._expand_and_sort_child_rows = function(rowItem, sortPropertyName, sortFunction, sortDir) {
    return rowItem.getChildren();
};

TreeView.prototype.toggleRowChecked = function(rowNum) {
    var row = this._rows[rowNum];
    var currCheck = row.isChecked;
    var newCheck = !currCheck;
    //dump("Row " + rowNum + " was " + currCheck + ", now  " + newCheck + "\n");
    this._rows[rowNum].isChecked = newCheck;
    var toggleRow = -1;
    if (row.isContainer) {
        widgets.repls.treeBoxObject.invalidateRow(rowNum);
        row.children.forEach(function(child) { child.isChecked = newCheck });
        toggleRow = rowNum;
    } else {
        // and update the children too, for when we collapse this tree
        var parentIndex = this.getParentIndex(rowNum);
        var container = this._rows[parentIndex];
        let children = container.children;
        var child = children[rowNum - parentIndex - 1];
        child.isChecked = newCheck;
        if (children.every(function(child) child.isChecked)) {
            container.isChecked = true;
        } else if (children.every(function(child) !child.isChecked)) {
            container.isChecked = false;
        }
        toggleRow = parentIndex;
    }
    if (toggleRow != -1) {
        // Do this in a setTimeout so all rows get invalidated.
        setTimeout(function() {
                try{
                let boxObject = widgets.repls.treeBoxObject;
                let topLine = boxObject.getFirstVisibleRow();
                this._wrapToggleOpenState(toggleRow);
                this._wrapToggleOpenState(toggleRow);
                boxObject.scrollToRow(topLine);
                updateUI('g_displayTree');
                } catch(ex) {
                    log.exception(ex, "Error toggling rows");
                }
            }.bind(this), 0);
    }
}
 
TreeView.prototype.toggleOpenState = function(rowNum) {
    this._invoked_toggleOpenState = true;
    return this._wrapToggleOpenState(rowNum);
};

TreeView.prototype._wrapToggleOpenState = function(rowNum) {
    return xtk.hierarchicalTreeView.prototype.toggleOpenState.apply(this, [rowNum]);
};

TreeView.prototype.getNextSiblingRow = function(idx, children) {
    if (this.isContainerOpen(idx)) {
        return idx + 1 + children.length;
    }
    return idx + 1;
};

var hitsByURL; /* { url => {
                     path => path,
                     hitListLines => { lineNum => { context: line of text,
                                                    hitList: [startIndex, endIndex,
                                                              columnNo]
                                                    }
                                      }
                         }
                    }
                    */


var filteredHitsByURL; // same as above 


var scopeStart;
var scopeEnd;
function rlcArray(a) {
    // run-length compress
    if (!a.length) {
        return [];
    }
    var x = [a[0].toString().substr(0, 40)];
    var lastVal = a[0];
    var numSkipped = 0;
    for (var idx = 1; idx < a.length; ++idx) {
        if (a[idx] == lastVal) {
            numSkipped++;
        } else {
            if (numSkipped) {
                x.push(numSkipped + " dups");
            }
            x.push(a[idx].toString().substr(0, 40))
            lastVal = a[idx];
            numSkipped = 0;
        }
    }
    if (numSkipped) {
        x.push(numSkipped + " dups");
    }
    return x;
}

function dumpHitList(label, aHitsByURL) {
    try {
    dump("dumpHitList: " + label + "\n");
    dump("keys aHitsByURL: "
         + Object.keys(aHitsByURL)
         + "\n");
    for (var url in aHitsByURL) {
        dump("url : " + url + "\n");
        if (!(url in aHitsByURL) || url == "*") {
            dump("No hits for url " + url + "\n");
            continue;
        }
        var hit = aHitsByURL[url];
        //dump(" keys(hit) = "
        //     + Object.keys(hit)
        //     + "\n");
        if (!('path' in hit)) {
            dump("No path for url " + url + "\n");
            continue;
        }
        dump(" file " + hit.path + ":\n");
        var match = g_defn.path == hit.path;
        var hitListLines = hit.hitListLines;
        for (let lineNum in hitListLines) {
            if (lineNum == "*") {
                continue;
            }
            dump("lineNum in hitListLines: " + lineNum + "\n");
            let hitListLineInfo = hitListLines[lineNum];
            if (!hitListLineInfo) {
                dump("No hitListLines[" + lineNum + "]\n");
                continue;
            }
            let inDefinitionContext = match && lineNum >= scopeStart && lineNum <= scopeEnd;
            let leadingChar = inDefinitionContext ? "*" : " ";
            let context = hitListLineInfo.context;
            let actualHitList = hitListLineInfo.hitList;
            for (let j = 0; j < actualHitList.length; j++) {
                let actualHit = actualHitList[j];
                //dump(" actualHitList["   + j + "]: " + actualHit[0] + "-"
                //     + actualHit[1] + " @" + actualHit[2] + "\n");
                dump(leadingChar);
                dump(" " + lineNum + ": " + actualHit[0] + "-" + actualHit[1] + " -- " + actualHit[2] + "\n");
                let actualColumn = actualHit[2] - 1;
                let amtBeforeStart = actualColumn < 10 ? 0 : actualColumn - 10;
                dump(leadingChar);
                dump(" context: "
                     + context.substring(amtBeforeStart, actualColumn)
                     + "**>"
                     + context.substring(actualColumn, actualColumn + actualHit[1] - actualHit[0])
                     + "<**"
                     + context.substr(actualColumn + actualHit[1] - actualHit[0], 10)
                    + "\n");
            }
        }
    }
    } catch(ex) {
        log.exception(ex, "Error in dumpHitList");
    }
}

function updateFilteredHits() {
    filteredHitsByURL = {};
    for (var url in hitsByURL) {
        if (!(url in hitsByURL)) {
            continue;
        }
        var hit = hitsByURL[url];
        if (!('path' in hit)) {
            continue;
        }
        filteredHitsByURL[url] = { path: hit.path, hitListLines: {} };
        var match = g_defn.path == hit.path;
        var hitListLines = hit.hitListLines;
        var filteredHitListLines = filteredHitsByURL[url].hitListLines;
        for (let lineNum in hitListLines) {
            //dump("lineNum in hitListLines: " + lineNum + "\n");
            let hitListLineInfo = hitListLines[lineNum];
            if (!hitListLineInfo) {
                //dump("No hitListLines[" + lineNum + "]\n");
                continue;
            }
            let inDefinitionContext = match && lineNum >= scopeStart && lineNum <= scopeEnd;
            let context = hitListLineInfo.context;
            let actualHitList = hitListLineInfo.hitList;
            filteredHitListLines[lineNum] = { context:context, hitList: [] };
            let filteredActualHitList = filteredHitListLines[lineNum].hitList;
            if (inDefinitionContext) {
                actualHitList.forEach(function(actualHit) {
                    filteredActualHitList.push([].concat(actualHit));
                });
            } else {
                actualHitList.forEach(function(actualHit) {
                    let contextBefore = context.substr(0, actualHit[2] - 1);
                    if (/\w+\.$/.test(contextBefore)) {
                        filteredActualHitList.push([].concat(actualHit));
                    }
                });
                if (!filteredActualHitList.length) {
                    delete filteredHitListLines[lineNum];
                }
            }
        }
        if (Object.keys(hitListLines).length == 0) {
            delete filteredHitsByURL[url];
        }
    }
}
        
function ResultsManagerView() {
    this.url_dirty_status = {};
}
ResultsManagerView.prototype = {
    Clear: function() {
        log.debug(">> ResultsManagerView.Clear\n");
    },
    AddFindResult: function() {
        log.debug(">> ResultsManagerView.AddFindResult\n");
    },
    AddReplaceResult: function() {
        log.debug(">> ResultsManagerView.AddReplaceResult\n");
    },
    AddFindResults: function(type_count, types,
                             url_count, urls,
                             si_count, startIndexs,
                             ei_count, endIndexs,
                             value_count, values,
                            fn_count, fileNames,
                            ln_count, lineNums,
                            cn_count, columnNums,
                            context_count, contexts) {
        try {
        log.debug(">> AddFindResults\n");
        log.debug("types: " + rlcArray(types)
             + ", urls: " + rlcArray(urls)
             + ", startIndexs" + rlcArray(startIndexs)
             + ", endIndexs" + rlcArray(endIndexs)
             + ", values" + rlcArray(values)
             + ", lineNums" + rlcArray(lineNums)
             + ", columnNums" + rlcArray(columnNums)
             + ", contexts" + rlcArray(contexts)
             + "\n");
        for (var i = 0; i < startIndexs.length; i++) {
            var url = urls[i];
            if (!(url in this.url_dirty_status)) {
                var views = g_opener.ko.views.manager.topView.
                    getViewsByTypeAndURI(true, "editor", url);
                if (views && views.length > 0) {
                    this.url_dirty_status[url] = views[0].isDirty;
                }
            }
            if (this.url_dirty_status[url]) {
                continue;
            }
            var hit;
            if (!(url in hitsByURL)) {
                log.debug("Add url "
                     + url
                     + ", filename: "
                     + fileNames[i]
                     + "\n");
                hitsByURL[url] = { path: fileNames[i], hitListLines: {} };
            }
            let hitListLines = hitsByURL[url].hitListLines;
            let lineNum = lineNums[i];
            if (!(lineNum in hitListLines)) {
                hitListLines[lineNum] = { context: contexts[i], hitList: [] };
            }
            log.debug("Add hit " + lineNum
                 + " -- " + startIndexs[i]
                 + ":" + endIndexs[i]
                 + " col " + columnNums[i]
                 + "\n");
            hitListLines[lineNum].hitList.push([startIndexs[i], endIndexs[i],
                                                columnNums[i]]);
        }
        } catch(ex) {
            log.debug("Error: " + ex + "\n");
        }
    },
    AddReplaceResults: function() {
        log.debug(">> ResultsManagerView.AddReplaceResults\n");
    },
    GetType:function() {
        log.debug(">> ResultsManagerView.GetType\n");
        return "type";
    },
    GetUrl:function() {
        log.debug(">> ResultsManagerView.GetUrl\n");
        return "url";
    },
    GetStartIndex:function() {
        log.debug(">> ResultsManagerView.GetStartIndex\n");
        return "StartIndex";
    },
    GetEndIndex:function() {
        log.debug(">> ResultsManagerView.GetEndIndex\n");
        return "EndIndex";
    },
    GetValue:function() {
        log.debug(">> ResultsManagerView.GetValue\n");
        return "value";
    },
    GetReplacement:function() {
        log.debug(">> ResultsManagerView.GetReplacement\n");
        return "replacement";
    },
    GetLineNum:function() {
        log.debug(">> ResultsManagerView.GetLineNum\n");
        return 0;
    },
    GetColumnNum:function() {
        log.debug(">> ResultsManagerView.GetColumnNum\n");
        return 0;
    },
    Sort:function() {
        log.debug(">> ResultsManagerView.Sort\n");
    },
    SetFilterText:function(val) {
        log.debug(">> ResultsManagerView.SetFilterText\n");
    },
    GetNumUrls:function() {
        log.debug(">> ResultsManagerView.GetNumUrls\n");
        return 0;
    },
    
    QueryInterface: function(iid) {
        if (!iid.equals(Ci.koIFindResultsView) &&
            !iid.equals(Ci.nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    }
};
function ResultsManager() {
    this.view = new ResultsManagerView();
}
ResultsManager.prototype = {
    searchStarted: function() {
        log.debug(">> ResultsManager.searchStarted\n");
    },
    searchFinished: function() {
        log.debug(">> ResultsManager.searchFinished\n");
        //g_hitsByURL = hitsByURL;
        //dumpHitList("raw hits:", hitsByURL);
        let url_dirty_status = this.view.url_dirty_status;
        let urls = Object.keys(url_dirty_status);
        let dirty_urls = urls.filter(function(url) url_dirty_status[url]);
        if (dirty_urls.length) {
            alert("Ignoring changes in file"
                  + (dirty_urls.length > 1 ? "s " : " ")
                  + dirty_urls.map(function(url) g_opener.ko.uriparse.URIToLocalPath(url)).join(", "));
        }
                  
            
        this.view.url_dirty_status = {};
        updateFilteredHits();
        //dumpHitList("filtered hits:", filteredHitsByURL);
        // Populate the tree, but not in the calling context from the
        // find-all async thing.
        setTimeout(function() {
                try {
                    doHits(filteredHitsByURL);
                } catch(ex) {
                    log.exception(ex, "Error in doHits");
                }
            }, 0);
    },
    setDescription: function(desc, important) {
        log.debug(">> ResultsManager.setDescription: important: "
             + important
             + ", desc:["
             + desc
             + "]\n");
    },
    configure: function(pattern, patternAlias, repl, context, options) {
        this._pattern = pattern;
        this._patternAlias = patternAlias;
        this._repl = repl;
        this.context_ = context;
        this._options = options;
    },
    id: 59,
    QueryInterface: function(iid) {
        if (!iid.equals(Ci.koIFindResultsTabManager) &&
            !iid.equals(Ci.nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    }
}

function startSearch() {
    hitsByURL = {};
    try {
    var context = Cc["@activestate.com/koFindInFilesContext;1"]
        .createInstance(Ci.koIFindInFilesContext);
    context.type = Ci.koIFindContext.FCT_IN_FILES;
    context.cwd = g_opener.ko.uriparse.dirName(g_defn.path);
        
    var find_svc = Components.classes["@activestate.com/koFindService;1"]
                      .getService(Components.interfaces.koIFindService);
    var opts = find_svc.options;
    opts.multiline = false;
    opts.patternType = Ci.koIFindOptions.FOT_SIMPLE;
    opts.caseSensitivity = g_args.matchCase;
    opts.matchWord = true;
    opts.encodedFolders = [];
    opts.searchInSubfolders = g_args.search_in_subdirs;
    opts.encodedIncludeFiletypes = g_args.includes;
    opts.encodedExcludeFiletypes = g_args.excludes;
    opts.showReplaceAllResults = true;
    opts.confirmReplacementsInFiles = true;
    var resultsMgr = new ResultsManager();
    var pattern = g_args.pattern;
    var patternAlias = pattern;
    resultsMgr.configure(pattern, patternAlias, null, context, opts);
        find_svc.findallinfiles(resultsMgr.id, pattern, resultsMgr);
    } catch (ex) {
        log.exception(ex, "startSearch/findallinfiles failed; ");
    }
}

function _apply_checked_changes(url, tree, rows, idx) {
    var koFS = Cc["@activestate.com/koFileService;1"].getService(Ci.koIFileService);
    var koIFileEx = koFS.getFileFromURI(url);
    koIFileEx.open("rb");
    var oldText = koIFileEx.readfile();
    koIFileEx.close();
    var oldTextLines = oldText.split(/(\r?\n|\r)/);
    var madeChange = false;
    let children = rows[idx].getChildren();
    let lim = children.length;
    let lastLineNo = 0;
    //dump("Got " + lim + " children\n");
    let idx = 0;
    for (; idx < lim; ++idx) {
        let child = children[idx];
        //dump("children[" + idx + "]: " + child.isChecked + "\n");
        if (!child.isChecked) {
            // No need to grab
            //dump("!child.isChecked: continue...\n");
            continue;
        }
        let thisLineNo = child.lineNo - 1;
        oldTextLines[thisLineNo * 2] = child.lineChangeMgr.getText();
        madeChange = true;
        // Skip further occurrences
        while (idx < lim - 1 && children[idx + 1].inLineOccurrence > 0) {
            idx++;
        }
    }
    if (madeChange) {
        koIFileEx.open("wb");
        koIFileEx.puts(oldTextLines.join(""));
        return url;
    }
    return null;
}

function _diff_checked_changes(url, tree, rows, idx) {
    var koOS = Cc["@activestate.com/koOs;1"].getService(Ci.koIOs);
    var oldText = koOS.readfile(url);
    var oldTextLines = oldText.split(/(\r?\n|\r)/);
    // Allow last line to not end with a newline.
    let newTextLines = [];
    let oldLineIdx = 0;
    let children = rows[idx].getChildren();
    let lim = children.length;
    let lastLineNo = 0;
    //dump("Got " + lim + " children\n");
    let idx = 0;
    for (; idx < lim; ++idx) {
        let child = children[idx];
        //dump("children[" + idx + "]: " + child.isChecked + "\n");
        if (!child.isChecked) {
            // No need to grab
            //dump("!child.isChecked: continue...\n");
            continue;
        }
        let thisLineNo = child.lineNo - 1;
        //dump("thisLineNo: " + thisLineNo + ", lastLineNo:" + lastLineNo + "\n");
        if (lastLineNo < thisLineNo) {
            newTextLines.push(oldTextLines.slice(lastLineNo * 2,
                                                 thisLineNo * 2).join(""));
        }
        lastLineNo = thisLineNo + 1;
        newTextLines.push(child.lineChangeMgr.getText());
        // Keep the old newline
        newTextLines.push(oldTextLines[thisLineNo * 2 + 1]);
        // Skip further occurrences
        while (idx < lim - 1 && children[idx + 1].inLineOccurrence > 0) {
            idx++;
        }
    }
    newTextLines.push(oldTextLines.slice(lastLineNo * 2).join(""));
    let koFileService = Cc["@activestate.com/koFileService;1"].
        createInstance(Ci.koIFileService);
    let outputTmpFile = koFileService.makeTempFile(g_opener.ko.uriparse.ext(url),
                                              "wb");
    try {
        outputTmpFile.puts(newTextLines.join(""));
    } finally {
        outputTmpFile.close();
    }
    // Verify outputTmpFile.path is deleted
    let newPath = outputTmpFile.path;
    let udiff = Cc["@activestate.com/koDiffService;1"].
                     getService(Ci.koIDiffService).diffFilepaths(url, newPath);
    //dump("diff (" + url
    //     + ", " + newPath + ") => ]" + udiff + "[\n");
    try {
        koFileService.deleteTempFile(newPath, true);
    } catch(ex) {
        log.exception("Problem deleting temp file");
    }
    return udiff;
}
    
function walk_loaded_paths(callback, context) {
    try {
        //dump(">> show_marked_changes\n");
        var tree = g_displayTree;
        var rows = tree._rows;
        var idx = 0, lim = rows.length;
        var children;
        var results = [];
        while (idx < lim) {
            let row = rows[idx];
            let url = row.path;
            //dump("idx: " + idx + "");
            let res = callback(url, tree, rows, idx);
            //dump(" ===> " + diff + "\n\n");
            if (res) {
                results.push(res);
            }
            idx = tree.getNextSiblingRow(idx, row.children);
        }
    } catch(ex) {
        log.exception(ex, "walk_loaded_paths(" + context + ")");
    }
    return results;
}

function make_changes() {
    var files = walk_loaded_paths(_apply_checked_changes, "in make_marked_changes");
    if (files.length) {
        g_opener.ko.statusBar.AddMessage("Updated " + files.length + " files", "editor", 3000);
    }
    return true;
}

function show_marked_changes() {
    var diffs = walk_loaded_paths(_diff_checked_changes,
                                  "in show_marked_changes");
    if (!diffs.length) {
        log.debug("No diffs selected\n");
    } else {
        ko.launch.diff(diffs.join("\n"), "Marked Changes");
    }
}

function updateUI(why) {
    if (typeof(why) == "undefined") why = null;
    if (why == 'g_displayTree' || !why) {
        let tree = g_displayTree;
        let rows = tree._rows;
        let idx = 0, lim = rows.length;
        let disableButton = true;
        while (idx < lim) {
            let row = rows[idx];
            let children = row.getChildren();
            if (row.isChecked
                || children.some(function(child) child.isChecked)) {
                disableButton = false;
                break;
            }
            idx = tree.getNextSiblingRow(idx, children);
        }
        widgets.showChangesButton.disabled = disableButton;
    }
}
                
            
