/* Copyright (c) 2000-2011 ActiveState Software Inc.
   See the file LICENSE.txt for licensing information. */

if (typeof(ko) == 'undefined') {
    var ko = {};
}

if (typeof(ko.perldevkit)!='undefined') {
    ko.logging.getLogger('').warn("ko.perldevkit was already loaded, re-creating it.\n");
}
ko.perldevkit = {};

(function() {

    /* locals */

    function PDKInfo(primaryPrefName, fallbackPrefName, primaryExeName, fallbackExeName) {
        this.log = ko.logging.getLogger("PDK");
        //this.log.setLevel(ko.logging.LOG_DEBUG);
        this.log.debug("PDKItemsController init");
        this.primaryPrefName = primaryPrefName;
        this.primaryExeName = primaryExeName;
        this.fallbackPrefName = fallbackPrefName;
        this.fallbackExeName = fallbackExeName;
        this._cache = {};
        try {
            this.infoService = Components.classes['@activestate.com/koInfoService;1'].
                    getService(Components.interfaces.koIInfoService);
            this.isWindows = this.infoService.platform.substring(0,3) == "win";

            this.osPathSvc = Components.classes["@activestate.com/koOsPath;1"].
                    getService(Components.interfaces.koIOsPath);
            this.koDirSvc = Components.classes["@activestate.com/koDirs;1"].
                    getService(Components.interfaces.koIDirs);
            this.perlInfo = Components.classes["@activestate.com/koAppInfoEx?app=Perl;1"].
                    getService(Components.interfaces.koIPerlInfoEx);
            this.sysUtils = Components.classes["@activestate.com/koSysUtils;1"].
                getService(Components.interfaces.koISysUtils);
            this.infoSvc = Components.classes["@activestate.com/koInfoService;1"].
                              getService(Components.interfaces.koIInfoService);
            this.prefs = Components.classes["@activestate.com/koPrefService;1"].
                        getService(Components.interfaces.koIPrefService).prefs;
            this.prefs.prefObserverService.addObserver(this, primaryPrefName, false);
        } catch(e) {
            this.log.exception(e);
        }
    }

    PDKInfo.prototype = {
        getLocation: function() {
            var location = null;
            if (typeof(this._cache['location']) != 'undefined') {
                return this._cache['location'];
            }
            try {
                if (this.prefs.hasStringPref(this.primaryPrefName)) {
                    location = this.prefs.getStringPref(this.primaryPrefName);
                }
                else if (this.prefs.hasStringPref(this.fallbackPrefName)) {
                    location = this.prefs.getStringPref(this.fallbackPrefName);
                }
                if (!location) {
                    var availInstalls = this.installationPaths();
                    if (availInstalls.length > 0)
                        location = availInstalls[0];
                }
                if (!location) return null;
            } catch(e) {
                this.log.exception(e);
            }
            this._cache['location'] = location;
            return location;
        },

        getAppPath: function(app) {
            if (typeof(this._cache[app]) != 'undefined' &&
                typeof(this._cache[app]['path']) != 'undefined') {
                return this._cache[app]['path'];
            }
            try {
                this._cache[app] = {};
                this._cache[app]['path'] = this.appPath(this.getLocation(), app);
                return this._cache[app]['path'];
            } catch(e) {
                this.log.exception(e);
            }
            return null;
        },

        appPath: function(location, app) {
            if (!location) return null;
            try {
                // given an app name (eg. perlapp), build the path
                if (this.isWindows) {
                    app += ".exe";
                }
                return this.osPathSvc.joinlist(3, [location, "bin", app]);
            } catch(e) {
                this.log.exception(e);
            }
            return null;
        },

        getPAIPath: function()
        {
            try {
                return this.PAIPath(this.getLocation());
            } catch(e) {
                this.log.exception(e);
            }
            return null;
        },

        PAIPath: function(location)
        {
            if (!location) return null;
            try {
                if (this.isWindows) {
                    return this.osPathSvc.joinlist(4, [location, "bin", "lib", "pai.exe"]);
                }
                return this.osPathSvc.joinlist(3, [location, "lib", "pai"]);
            } catch(e) {
                this.log.exception(e);
            }
            return null;
        },

        hasApp: function (app) {
            try {
                if (typeof(this._cache[app]) != 'undefined' &&
                    typeof(this._cache[app]['exists']) != 'undefined' &&
                    this._cache[app]['exists']) {
                    return true;
                }
                var app_path = this.getAppPath(app);
                if (!app_path) return false;
                this._cache[app]['exists'] =  this.osPathSvc.exists(app_path);
                return this._cache[app]['exists'];
            } catch(e) {
                this.log.exception(e);
            }
            return false;
        },

        exists: function (location) {
            try {
                return this.osPathSvc.exists(this.PAIPath(location));
            } catch(e) {
                this.log.exception(e);
            }
            return false;
        },

        _BaseDir: function(location)
        {
            // pdk puts executables in pdkdir/bin on all platforms
            // pdkDir/bin/someapp.exe
            try {
                return this.osPathSvc.dirname(this.osPathSvc.dirname(location));
            } catch(e) {
                this.log.exception(e);
            }
            return null;
        },

        installationPaths: function()
        {
            var availInterps = new Array();
            try {
                var found = this.sysUtils.WhichAll(this.primaryExeName, new Object());
                var found2 = this.sysUtils.WhichAll(this.fallbackExeName, new Object());
                found = found.concat(found2);
                var dir = null;
                for (var i=0; i < found.length; i++) {
                    if (found[i]) {
                        availInterps.push(this._BaseDir(found[i]));
                    }
                }
                if (this.infoSvc.platform == 'darwin') {
                    // bug 80816: hardwire these because they're hard to find
                    var installDir = '/usr/local/PDK';
                    var installDirBin = installDir + "/bin";
                    if ((this.osPathSvc.exists(this.osPathSvc.join(installDirBin, this.primaryExeName))
                         || this.osPathSvc.exists(this.osPathSvc.join(installDirBin, this.fallbackExeName)))
                        && availInterps.indexOf(installDir) == -1) {
                        availInterps.push(installDir);
                    }
                }
            } catch(ex) {
                log.exception(ex);
            }
            return availInterps;
        },

        hasLicense: function() {
            try {
                if (typeof(this._cache['license']) != 'undefined' &&
                    this._cache['license']) return true;

                var interp = ko.prefs.getStringPref("perlDefaultInterpreter");
                if (interp == '') {
                    interp = 'perl';
                    interp = this.sysUtils.Which(interp);
                    if (!interp) {
                        ko.dialogs.alert("This feature requires that you have " +
                              "Perl on your PATH or that the Perl default " +
                              "interpreter preference be correctly " +
                              "set, and that a licensed copy of Perl " +
                              "Dev Kit 6.0 or later be installed in " +
                              "the same directory.");
                    }
                }
                var perlInfoEx = Components.classes["@activestate.com/koAppInfoEx?app=Perl;1"].createInstance(Components.interfaces.koIPerlInfoEx);
                perlInfoEx.installationPath = perlInfoEx.getInstallationPathFromBinary(interp);
                var sysDataSvc = Components.classes["@activestate.com/koSystemDataService;1"]
                                 .getService(Components.interfaces.koISystemDataService);

                var perlBinDir = null;
                if (ko.prefs.hasStringPref(this.primaryPrefName)) {
                    // hack in the pdkLocation for the old pdk dialog, this enables
                    // support of different installations in the old dialog
                    perlBinDir = ko.prefs.getStringPref(this.primaryPrefName);
                    if (perlBinDir) {
                        perlBinDir = this.osPathSvc.join(perlBinDir,'bin');
                    }
                }
                if (!perlBinDir) {
                    // is pdk 6 on the path?
                    var paths = this.installationPaths();
                    if (paths.length) {
                        perlBinDir = this.osPathSvc.join(paths[0],'bin');
                    } else {
                        perlBinDir = this.osPathSvc.dirname(interp);
                    }
                }
                sysDataSvc.setString("pdk.perlbinpath", perlBinDir);

                if (! sysDataSvc.getString('pdk.perlapp')) {
                    ko.dialogs.alert("Komodo could not find an installation of Perl Dev Kit in: '" + perlBinDir + "'.");
                    return false;
                }
                if (! sysDataSvc.getBoolean("pdk.haveLicense")) {
                    ko.dialogs.alert("This feature requires that you have " +
                          "a license for the Perl Dev Kit installed.");
                    return false;
                }
                var version = parseFloat(sysDataSvc.getString("pdk.version"));
                if (!version) {
                    ko.dialogs.alert("Komodo could not determine the version number for Perl Dev Kit at ["+perlBinDir+"].");
                    return false;
                }
                if (version < 6) {
                    ko.dialogs.alert("Komodo requires Perl Dev Kit version 6 or later.");
                    return false;
                }
                this._cache['license'] = true;
                return true;
            } catch (e) {
                log.error(e);
            }
            return false;
        },

        observe: function(subject, topic, data)
        {
            this.log.debug('observe '+topic);
            try {
                switch (topic) {
                case this.primaryPrefName:
                    this._cache = {};
                    if (topic == "pdkLocation") {
                        // Needs setTimeout, otherwise the setBooleanPref will
                        // fail to take... a bug in prefs?
                        window.setTimeout(UpdateToolbarVisibility, 1);
                    }
                    window.updateCommands("pdk_prefchange");
                }
            } catch(e) {
                this.log.exception(e);
            }
        }

    }

    function UpdateToolbarVisibility() {
        var available = ko.perldevkit.info.getLocation();
        if (available && (!ko.prefs.hasPref("pdk_installed") ||
                          !ko.prefs.getBooleanPref('pdk_installed'))) {
            document.getElementById('tb_pdkPerlMenu').removeAttribute('kohidden');
            ko.prefs.setBooleanPref('pdk_installed', 1);
        } else if (!available && ko.prefs.getBoolean('pdk_installed', false)) {
            document.getElementById('tb_pdkPerlMenu').setAttribute('kohidden', 'true');
            ko.prefs.setBooleanPref('pdk_installed', 0);
        }
    }

    this.initialize = function PDKInfoInit() {
        if (!ko.perldevkit.info) {
            ko.perldevkit.info = new PDKInfo('pdkLocation','pdkGuiLocation','perlapp','perlfb');
            ko.perldevkit.guiInfo = new PDKInfo('pdkGuiLocation','pdkLocation','perlfb','perlapp');
            
            var el = document.getElementById('popup_pdk');
            // popup_pdk will only exist in the top level window.  When the
            // prefs window uses pdkinfo, this will fail.
            if (!el) return
            
            // Hide apps not on unix.
            if (!ko.perldevkit.info.isWindows) {
                document.getElementById('menu_pdkPerlSvc').setAttribute('collapsed', 'true');
                document.getElementById('menu_pdkPerlCtrl').setAttribute('collapsed', 'true');
                document.getElementById('menu_pdkPerlNET').setAttribute('collapsed', 'true');
                document.getElementById('menu_pdkPerlTray').setAttribute('collapsed', 'true');
                document.getElementById('menu_pdkPerlVBConverter').setAttribute('collapsed', 'true');
                document.getElementById('tb_pdkPerlSvc').setAttribute('collapsed', 'true');
                document.getElementById('tb_pdkPerlCtrl').setAttribute('collapsed', 'true');
                document.getElementById('tb_pdkPerlNET').setAttribute('collapsed', 'true');
                document.getElementById('tb_pdkPerlTray').setAttribute('collapsed', 'true');
                document.getElementById('tb_pdkPerlVBConverter').setAttribute('collapsed', 'true');
            }
            UpdateToolbarVisibility();
        }
    }

}).apply(ko.perldevkit);

/**
 * @deprecated since 7.0
 */
ko.logging.globalDeprecatedByAlternative('gPDKInfo', 'ko.perldevkit.info');
ko.logging.globalDeprecatedByAlternative('gPDKGuiInfo', 'ko.perldevkit.guiInfo');
ko.logging.globalDeprecatedByAlternative('PDKInfoInit', 'ko.perldevkit.initialize');

